import yaml
import os

file_dir = os.path.dirname(__file__)

class Config:
    def __init__(self, buy_params, wallet_data, general_params):
        self.buy_params = buy_params
        self.wallet_data = wallet_data
        self.general_params = general_params

def get_config(buy_params_path=None, wallet_data_path=None, general_params_path=None):

    buy_params_path = f'{file_dir}/buy_params.yml' if buy_params_path is None else buy_params_path
    wallet_data_path = f'{file_dir}/wallet_data_template.yml' if wallet_data_path is None else wallet_data_path
    general_params_path = f'{file_dir}/general_params.yml' if general_params_path is None else general_params_path

    with open(buy_params_path, 'r') as f:
        buy_params = yaml.safe_load(f)

    with open(wallet_data_path, 'r') as f:
        wallet_data = yaml.safe_load(f)

    with open(general_params_path, 'r') as f:
        general_params = yaml.safe_load(f)

    return Config(buy_params, wallet_data, general_params)


import abc
import json
import time

from eth_account import messages
from web3 import Web3
from dexbuytools import log_utils
from dexbuytools.helpers.data.general import ERC20_ABI
from dexbuytools.helpers.data.eth import chain_data as eth_chain_data
from eth_account.account import Account
import requests


class EvmBaseHelper(abc.ABC):

    def __init__(self, w3, chain_data, dex_name, config):
        self.w3 = w3
        self.chain_data = chain_data
        self.config = config
        self.dex_router = self._get_dex_router_contract(dex_name)

    @abc.abstractmethod
    def buy_instantly(self, token_address):
        pass

    def build_uniswapv2_style_tx(self, token_address, wallet_address, swap_method='swapExactETHForTokensSupportingFeeOnTransferTokens'):
        token_address = self.w3.toChecksumAddress(token_address)
        main_token_address = self.w3.toChecksumAddress(self.chain_data['MAIN_TOKEN_ADDRESS'])
        amount_out_min = self.config.buy_params['AMOUNT_OUT_MIN']
        path = [main_token_address, token_address]
        to = wallet_address
        deadline = round(time.time()) + self.config.buy_params['DEADLINE_OFFSET']

        if self.config.buy_params['USE_WRAPPED_MAIN_TOKEN']:
            swap_method = 'swapExactTokensForTokensSupportingFeeOnTransferTokens'
            amount_in = int(self.config.buy_params['AMOUNT'] * 10 ** 18)
            tx = self.dex_router.functions[swap_method](
                amount_in,
                amount_out_min,
                path,
                to,
                deadline
            ).buildTransaction({
                'nonce': self.w3.eth.getTransactionCount(wallet_address),
                'gas': 3000000,
                'gasPrice': self.w3.eth.gas_price + self.config.buy_params['GAS_PRICE_BONUS']
            })
        else:
            tx = self.dex_router.functions[swap_method](
                amount_out_min,
                path,
                to,
                deadline
            ).buildTransaction({
                'value': int(self.config.buy_params['AMOUNT'] * 10 ** 18),
                'nonce': self.w3.eth.getTransactionCount(wallet_address),
                'gasPrice': self.w3.eth.gas_price + self.config.buy_params['GAS_PRICE_BONUS']
            })

            # for some reason estimating gas does not work for method 'swapExactTokensForTokensSupportingFeeOnTransferTokens'
            gas_estimate = self.w3.eth.estimate_gas(tx)
            if gas_estimate * tx['gasPrice'] > self.config.buy_params["MAX_GAS_ESTIMATE"] * 10 ** 18:
                log_utils.log_error("Transaction not executed, because the gas estimate is greater than the limit "
                                    f"configured in the buy_params-file. Gas estimate in wei : {tx['gas'] * tx['gasPrice']}")
                raise NotImplementedError

            tx['gas'] = gas_estimate + self.config.buy_params['GAS_LIMIT_BONUS']

        return tx

    def perform_uniswapv2_style_buy(
            self,
            token_address):

        wallet_address = self.w3.toChecksumAddress(self._get_wallet_address_from_key())

        tx = self.build_uniswapv2_style_tx(token_address, wallet_address)

        signed_tx = self.w3.eth.account.signTransaction(tx, private_key=self.config.wallet_data['PRIVATE_KEY'])
        tx_hash = self.w3.eth.sendRawTransaction(signed_tx.rawTransaction)
        receipt = self.w3.eth.waitForTransactionReceipt(tx_hash)

        log_utils.log_info(f"buy of {token_address} performed on DEX {self.dex_router.address}. receipt: {receipt}")

        if self.config.buy_params['APPROVE_AFTER_BUY']:
            self._approve(self.w3, self.dex_router.address, token_address, wallet_address)

        return receipt

    def _approve(self, w3, dex_address, token_address, wallet_address):
        erc20_token = w3.eth.contract(
            abi=ERC20_ABI,
            address=token_address
        )

        amount_to_approve = erc20_token.functions.balanceOf(wallet_address).call()

        tx = erc20_token.functions.approve(dex_address, amount_to_approve).buildTransaction({
            'nonce': w3.eth.getTransactionCount(wallet_address),
            'gas': 300000,  # arbitrary
            'gasPrice': w3.eth.gas_price + self.config.buy_params['GAS_PRICE_BONUS']
        })

        signed_tx = w3.eth.account.signTransaction(tx, private_key=self.config.wallet_data['PRIVATE_KEY'])
        tx_hash = w3.eth.sendRawTransaction(signed_tx.rawTransaction)
        receipt = w3.eth.waitForTransactionReceipt(tx_hash)

        log_utils.log_info(f"approval given for dex {dex_address} to spend {amount_to_approve} of {token_address}. "
                           f"receipt: {receipt}")

    def _get_wallet_address_from_key(self):
        return self.w3.eth.account.from_key(self.config.wallet_data['PRIVATE_KEY']).address

    @abc.abstractmethod
    def buy_on_liquidity(self, buy_params, address=None, search_name=None, search_symbol=None):
        pass

    def _get_dex_router_contract(self, dex_name):
        return self.w3.eth.contract(
            abi=self.chain_data[f"ROUTER_ABI_{dex_name}"],
            address=self.chain_data[f"ROUTER_ADDRESS_{dex_name}"]
        )


import json

import requests
import web3
from eth_account import messages, Account
from web3 import Web3
from web3.middleware import geth_poa_middleware

from dexbuytools import log_utils
from dexbuytools.helpers.EvmBaseHelper import EvmBaseHelper
from dexbuytools.helpers.data.eth import chain_data
from uniswap import Uniswap
from dexbuytools.helpers.data.eth import chain_data as eth_chain_data


class EthHelper(EvmBaseHelper):
    DEFAULT_DEX = "UNI"

    def __init__(self, config, dex_name=None, custom_rpc=None):
        w3 = Web3(Web3.HTTPProvider(config.general_params['ETH_RPC_URL'] if custom_rpc is None else custom_rpc))

        dex_name = EthHelper.DEFAULT_DEX if dex_name is None else dex_name
        super().__init__(w3, chain_data, dex_name, config)

    def buy_instantly(self, token_address):
        uniswap_v2 = Uniswap(address=None, private_key=None, version=2,
                             provider=self.config.general_params['ETH_RPC_URL'])
        uniswap_v3 = Uniswap(address=None, private_key=None, version=3,
                             provider=self.config.general_params['ETH_RPC_URL'])
        token_address = self.w3.toChecksumAddress(token_address)
        weth_address = self.w3.toChecksumAddress(self.chain_data['MAIN_TOKEN_ADDRESS'])

        output_v2 = 0
        output_v3 = 0

        try:
            output_v2 = uniswap_v2.get_price_input(weth_address, token_address,
                                                   int(self.config.buy_params["AMOUNT"] * 10 ** 18))
        except web3.exceptions.ContractLogicError as e:
            pass

        try:
            output_v3 = uniswap_v3.get_price_input(weth_address, token_address,
                                                   int(self.config.buy_params["AMOUNT"] * 10 ** 18))
        except web3.exceptions.ContractLogicError as e:
            pass

        if output_v2 == 0 and output_v3 == 0:
            log_utils.log_error(f"No liquidity on uniswap for token f{token_address}")
            return

        if output_v2 > output_v3:
            return self._perform_uniswapv2_buy(token_address)
        else:
            return self._perform_uniswapv3_buy(token_address)

    def _perform_uniswapv2_buy(self, token_address):
        wallet_address = self.w3.toChecksumAddress(self._get_wallet_address_from_key())
        tx = self.build_uniswapv2_style_tx(token_address, wallet_address)

        del tx['gasPrice']  # not required post EIP-1559

        latest_block = self.w3.eth.get_block('latest')
        tx['type'] = 2
        tx['chainId'] = 1
        tx['maxPriorityFeePerGas'] = 0  # send bribe directly with flashbots later on, hence no priority fee
        tx['maxFeePerGas'] = 2 * latest_block['baseFeePerGas']
        tx['nonce'] = self.w3.eth.getTransactionCount(wallet_address)

        self._send_flashbots_bundle(tx, wallet_address)

    def _perform_uniswapv3_buy(self, token_address):
        uniswapv3 = self.w3.eth.contract(
            abi=chain_data['ROUTER_ABI_UNIV3'],
            address=chain_data['ROUTER_ADDRESS_UNIV3']
        )

        log_utils.log_error("Buying on UniswapV3 not yet implemented")

    def buy_on_liquidity(self, buy_params, address=None, search_name=None, search_symbol=None):
        raise NotImplementedError()

    def _send_flashbots_bundle(self, tx, wallet_address):

        signed_tx = self.w3.eth.account.signTransaction(tx, private_key=self.config.wallet_data['PRIVATE_KEY'])

        flashbots_check_and_send_contract = self.w3.eth.contract(
            abi=eth_chain_data[f"FLASHBOTS_CHECK_AND_SEND_ABI"],
            address=eth_chain_data[f"FLASHBOTS_CHECK_AND_SEND_ADDRESS"]
        )
        latest_block = self.w3.eth.get_block('latest')
        bribe_tx = flashbots_check_and_send_contract.functions.checkBytesAndSendMulti([], [], []).buildTransaction({
            'nonce': self.w3.eth.getTransactionCount(wallet_address) + 1,
            'type': 2,  # EIP-1559
            'chainId': 1,
            'gas': 300000,  # =gasLimit; REVIEW: use params to set this?
            'value': int(self.config.buy_params['BRIBE'] * 10 ** 18),
            # bribe send with 'value' param
            'maxPriorityFeePerGas': 0,
            'maxFeePerGas': 2 * latest_block['baseFeePerGas']
        })

        signed_bribe_tx = self.w3.eth.account.signTransaction(bribe_tx, private_key=self.config.wallet_data['PRIVATE_KEY'])

        bundle = [self.w3.toHex(signed_tx.rawTransaction), self.w3.toHex(signed_bribe_tx.rawTransaction)]

        # FlashBotsUtil.simulate(bundle, Web3.toHex(self.w3.eth.block_number + 1), self.config.wallet_data['PRIVATE_KEY'])
        # pass

        block_number = self.w3.eth.block_number
        for i in range(1, 3):
            FlashBotsUtil.send_bundle(bundle, block_number+i, self.config.wallet_data['PRIVATE_KEY'])

class FlashBotsUtil:
    RELAY_URL = 'https://relay.flashbots.net/'

    @staticmethod
    def send_bundle(bundle, block_number, private_key):
        headers = {
            'Content-Type': 'application/json'
        }

        payload = json.dumps({
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'eth_sendBundle',
            'params': [
                {
                    'txs': bundle,
                    'blockNumber': block_number
                }
            ]
        })

        headers['X-Flashbots-Signature'] = FlashBotsUtil._get_flashbots_signature(private_key, payload)

        response = requests.post(FlashBotsUtil.RELAY_URL, data=payload, headers=headers)
        log_utils.log_info(f'Bundle sent to FlashBots. Response: {response.text}')

    @staticmethod
    def simulate(bundle, block_number, private_key):
        headers = {
            'Content-Type': 'application/json'
        }

        payload = json.dumps({
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'eth_callBundle',
            'params': [
                {
                    'txs': bundle,
                    'blockNumber': block_number,
                    'stateBlockNumber': 'latest'
                }
            ]
        })

        headers['X-Flashbots-Signature'] = FlashBotsUtil._get_flashbots_signature(private_key, payload)

        response = requests.post(FlashBotsUtil.RELAY_URL, data=payload, headers=headers)
        log_utils.log_info(f'Called for bundle simulation. Response: {response.text}')

    @staticmethod
    def _get_flashbots_signature(private_key, payload_json):
        message = messages.encode_defunct(text=Web3.keccak(text=payload_json).hex())
        return Account.from_key(private_key).address + ':' + Web3.toHex(Account.sign_message(message, private_key).signature)



from web3 import Web3
from web3.middleware import geth_poa_middleware

from dexbuytools.helpers.EvmBaseHelper import EvmBaseHelper
from dexbuytools.helpers.data.bsc import chain_data
import dexbuytools.log_utils as log_utils

class BscHelper(EvmBaseHelper):

    DEFAULT_DEX = 'PCS' #PancakeSwap

    def __init__(self, config, dex_name=None, custom_rpc=None):
        w3 = Web3(Web3.HTTPProvider(config.general_params['BSC_RPC_URL'] if custom_rpc is None else custom_rpc))
        w3.middleware_onion.inject(geth_poa_middleware, layer=0)

        dex_name = BscHelper.DEFAULT_DEX if dex_name is None else dex_name
        super().__init__(w3, chain_data, dex_name, config)

    def buy_instantly(self, token_address):
        return super().perform_uniswapv2_style_buy(token_address)

    def buy_on_liquidity(self, buy_params, address=None, search_name=None, search_symbol=None):
        raise NotImplementedError()

        #liquidity_details = super().listen_for_liquidity(w3,dex,...)
        # if xy: super().perform_uniswap_style_buy(...





from web3 import Web3
from web3.middleware import geth_poa_middleware

from dexbuytools import log_utils
from dexbuytools.helpers.EvmBaseHelper import EvmBaseHelper
from dexbuytools.helpers.data.poly import chain_data


class PolyHelper(EvmBaseHelper):

    DEFAULT_DEX = "QS"

    def __init__(self, config, dex_name=None, custom_rpc=None):
        w3 = Web3(Web3.HTTPProvider(config.general_params['POLY_RPC_URL'] if custom_rpc is None else custom_rpc))
        w3.middleware_onion.inject(geth_poa_middleware, layer=0)

        dex_name = PolyHelper.DEFAULT_DEX if dex_name is None else dex_name
        super().__init__(w3, chain_data, dex_name, config)

    def buy_instantly(self, token_address):
        return super().perform_uniswapv2_style_buy(token_address)


    def buy_on_liquidity(self, buy_params, address=None, search_name=None, search_symbol=None):
        raise NotImplementedError()


